/*
 * Timer/counter unit support.
 *
 * Copyright (C) 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>
#include <l4/sys/err.h>
#include <l4/sys/irq.h>
#include <l4/sys/rcv_endpoint.h>
#include <l4/util/util.h>

#include <systypes/thread.h>

#include <math.h> // log2

#include "tcu-common.h"



// Register locations.

enum Regs : unsigned
{
  Tcu_enable_status                 = 0x010,      // TER
  Tcu_set_enable                    = 0x014,      // TESR
  Tcu_clear_enable                  = 0x018,      // TECR
  Tcu_stop_status                   = 0x01c,      // TSR
  Tcu_set_stop                      = 0x02c,      // TSSR
  Tcu_clear_stop                    = 0x03c,      // TSCR
  Tcu_flag_status                   = 0x020,      // TFR
  Tcu_set_flag                      = 0x024,      // TFSR
  Tcu_clear_flag                    = 0x028,      // TFCR
  Tcu_mask_status                   = 0x030,      // TMR
  Tcu_set_mask                      = 0x034,      // TMSR
  Tcu_clear_mask                    = 0x038,      // TMCR

  // Channel-related locations.

  Tcu_full_data_value_base          = 0x040,      // TDFRn
  Tcu_half_data_value_base          = 0x044,      // TDHRn
  Tcu_counter_base                  = 0x048,      // TCNTn
  Tcu_control_base                  = 0x04c,      // TCRn

  // Block size/step/offset for the above register set.

  Tcu_data_block_offset             = 0x010,
};

// Field definitions.

// Enable/stop register bits.

enum Channel_bit_numbers : unsigned
{
  Channel_wdt                       = 16,         // WDTS only

  // Enable/stop/flag/mask bit numbers.

  Channel_ost                       = 15,         // OSTEN/OSTS/OSTFLAG
  Channel_tcu7                      = 7,          // TCEN7/STOP7/FFLAG7/SFLAG7
  Channel_tcu6                      = 6,          // TCEN6/STOP6/FFLAG6/SFLAG6
  Channel_tcu5                      = 5,          // TCEN5/STOP5/FFLAG5/SFLAG5
  Channel_tcu4                      = 4,          // TCEN4/STOP4/FFLAG4/SFLAG4
  Channel_tcu3                      = 3,          // TCEN3/STOP3/FFLAG3/SFLAG3
  Channel_tcu2                      = 2,          // TCEN2/STOP2/FFLAG2/SFLAG2
  Channel_tcu1                      = 1,          // TCEN1/STOP1/FFLAG1/SFLAG1
  Channel_tcu0                      = 0,          // TCEN0/STOP0/FFLAG0/SFLAG0
};

// Flag/mask register bits.

enum Flag_bit_numbers : unsigned
{
  Half_match_wdt                    = 24,         // HFLAGW

  // Flag/mask group bit offsets.

  Half_match_shift                  = 16,
  Full_match_shift                  = 0,
};

// Counter data constraints.

enum Data_masks : unsigned
{
  Data_mask                         = 0xffff,
};

enum Control_bits : unsigned
{
  Count_prescale_field_mask         = 0x7,        // PRESCALE
  Count_prescale_max                = 5,          // CLK/1024
  Count_prescale_field_shift        = 3,

  Count_clock_field_mask            = 0x7,
  Count_clock_exclk                 = 4,          // EXT_EN
  Count_clock_rtclk                 = 2,          // RTC_EN
  Count_clock_pclk                  = 1,          // PCK_EN
  Count_clock_field_shift           = 0,
};



// Channel abstraction.

Tcu_channel::Tcu_channel(l4_addr_t addr, uint8_t channel, l4_cap_idx_t irq)
: _channel(channel), _irq(irq)
{
  _regs = new Hw::Mmio_register_block<32>(addr);
}

// Utility methods.
// NOTE: Also defined in the CPM abstraction, should be consolidated.

uint32_t
Tcu_channel::get_field(uint32_t reg, uint32_t mask, uint8_t shift)
{
  return (_regs[reg] & (mask << shift)) >> shift;
}

void
Tcu_channel::set_field(uint32_t reg, uint32_t mask, uint8_t shift, uint32_t value)
{
  _regs[reg] = (_regs[reg] & (~(mask << shift))) | ((mask & value) << shift);
}

// Operation methods.

uint8_t
Tcu_channel::get_clock()
{
  return (uint8_t) get_field(Tcu_control_base + _channel * Tcu_data_block_offset,
                             Count_clock_field_mask, Count_clock_field_shift);
}

void
Tcu_channel::set_clock(uint8_t clock)
{

  set_field(Tcu_control_base + _channel * Tcu_data_block_offset,
            Count_clock_field_mask, Count_clock_field_shift, clock);
}

uint32_t
Tcu_channel::get_prescale()
{
  return 1UL << (2 * get_field(Tcu_control_base + _channel * Tcu_data_block_offset,
                               Count_prescale_field_mask, Count_prescale_field_shift));
}

void
Tcu_channel::set_prescale(uint32_t prescale)
{
  // Obtain the log4 value for prescale.

  uint32_t value = (uint32_t) log2(prescale) / 2;

  set_field(Tcu_control_base + _channel * Tcu_data_block_offset,
            Count_prescale_field_mask, Count_prescale_field_shift,
            value > Count_prescale_max ? Count_prescale_max : value);
}

void
Tcu_channel::disable()
{
  _regs[Tcu_clear_enable] = 1UL << _channel;
}

void
Tcu_channel::enable()
{
  _regs[Tcu_set_enable] = 1UL << _channel;
}

bool
Tcu_channel::is_enabled()
{
  return _regs[Tcu_enable_status] & (1UL << _channel);
}

uint32_t
Tcu_channel::get_counter()
{ 
  return _regs[Tcu_counter_base + _channel * Tcu_data_block_offset] & Data_mask;
}

void
Tcu_channel::set_counter(uint32_t value)
{
  _regs[Tcu_counter_base + _channel * Tcu_data_block_offset] = value & Data_mask;
}

uint8_t
Tcu_channel::get_count_mode()
{
  return 0;
}

void
Tcu_channel::set_count_mode(uint8_t mode)
{
  if (mode != 0)
    throw -L4_EINVAL;
}

uint32_t
Tcu_channel::get_full_data_value()
{
  return _regs[Tcu_full_data_value_base + _channel * Tcu_data_block_offset] & Data_mask;
}

void
Tcu_channel::set_full_data_value(uint32_t value)
{
  _regs[Tcu_full_data_value_base + _channel * Tcu_data_block_offset] = value & Data_mask;
}

uint32_t
Tcu_channel::get_half_data_value()
{
  return _regs[Tcu_half_data_value_base + _channel * Tcu_data_block_offset] & Data_mask;
}

void
Tcu_channel::set_half_data_value(uint32_t value)
{
  _regs[Tcu_half_data_value_base + _channel * Tcu_data_block_offset] = value & Data_mask;
}

bool
Tcu_channel::get_full_data_mask()
{ 
  return _regs[Tcu_mask_status] & (1UL << (_channel + Full_match_shift));
}

void
Tcu_channel::set_full_data_mask(bool masked)
{
  _regs[masked ? Tcu_set_mask : Tcu_clear_mask] = (1UL << (_channel + Full_match_shift));
}

bool
Tcu_channel::get_half_data_mask()
{ 
  return _regs[Tcu_mask_status] & (1UL << (_channel + Half_match_shift));
}

void
Tcu_channel::set_half_data_mask(bool masked)
{
  _regs[masked ? Tcu_set_mask : Tcu_clear_mask] = (1UL << (_channel + Half_match_shift));
}

// Wait indefinitely for an interrupt request, returning true if one was delivered.

bool
Tcu_channel::wait_for_irq()
{
  if (l4_error(l4_rcv_ep_bind_thread(_irq, get_current_thread(), 0)))
    return false;

  bool irq = !l4_error(l4_irq_receive(_irq, L4_IPC_NEVER)) && have_interrupt();

  if (irq)
    ack_irq();

  return irq;
}

// Wait up to the given timeout (in microseconds) for an interrupt request,
// returning true if one was delivered.

bool
Tcu_channel::wait_for_irq(unsigned int timeout)
{ 
  if (l4_error(l4_rcv_ep_bind_thread(_irq, get_current_thread(), 0)))
    return false;

  bool irq = !l4_error(l4_irq_receive(_irq, l4_timeout(L4_IPC_TIMEOUT_NEVER, l4util_micros2l4to(timeout)))) && have_interrupt();

  if (irq)
    ack_irq();

  return irq;
}

// Acknowledge an interrupt condition.

void
Tcu_channel::ack_irq()
{
  _regs[Tcu_clear_flag] = 1UL << _channel;
}

// Return whether an interrupt is pending on the given channel.

bool
Tcu_channel::have_interrupt()
{
  return _regs[Tcu_flag_status] & (1UL << _channel) ? true : false;
}



// Peripheral abstraction.

Tcu_chip::Tcu_chip(l4_addr_t start, l4_addr_t end)
: _start(start), _end(end)
{
}

// Obtain a channel object.

Tcu_channel *
Tcu_chip::get_channel(uint8_t channel, l4_cap_idx_t irq)
{
  if (channel < num_channels())
    return _get_channel(_start, channel, irq);
  else
    throw -L4_EINVAL;
}
